import configparser
import os
import sys
import textwrap
import tomllib
from typing import Any, Optional

from cargo2rpm.metadata import FeatureFlags
import jsonschema

from rust2rpm import log


TOML_SCHEMA = {
    "type": "object",
    "properties": {
        # package properties
        "package": {
            "type": "object",
            "properties": {
                # override package Summary
                "summary": {
                    "type": "string",
                },
                # override package description
                "description": {
                    "type": "string",
                },
                # override package project URL
                "url": {
                    "type": "string",
                },
                # override package source URL
                "source-url": {
                    "type": "string",
                },
                # supported target architectures
                "supported-arches": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                # override binary subpackage name
                "bin-package-name": {
                    "type": "string",
                },
                # override installation of binary targets
                "cargo-install-bin": {
                    "type": "boolean",
                },
                # override installation of library sources
                "cargo-install-lib": {
                    "type": "boolean",
                },
                # override default debuginfo level (2)
                "debuginfo-level": {
                    "type": "integer",
                    "enum": [0, 1, 2],
                },
                # comments for manual Cargo.toml patches (rust2rpm -p)
                "cargo-toml-patch-comments": {
                    "type": "array",
                    "items": {
                        "type": "string",
                    },
                },
                # additional source files
                "extra-sources": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "file": {
                                "type": "string",
                            },
                            "number": {
                                "type": "integer",
                            },
                            "comments": {
                                "type": "array",
                                "items": {
                                    "type": "string",
                                },
                            },
                        },
                        "additionalProperties": False,
                        "required": ["file", "number"],
                    },
                },
                # additional patch files
                "extra-patches": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "file": {
                                "type": "string",
                            },
                            "number": {
                                "type": "integer",
                            },
                            "comments": {
                                "type": "array",
                                "items": {
                                    "type": "string",
                                },
                            },
                        },
                        "additionalProperties": False,
                        "required": ["file", "number"],
                    },
                },
                # additional packaged files
                "extra-files": {
                    "type": "array",
                    "items": {
                        "type": "string",
                    },
                },
                # rename binaries to avoid conflicts
                "bin-renames": {
                    "type": "object",
                    "patternProperties": {
                        "^.*$": {
                            "type": "string",
                        },
                    },
                    "additionalProperties": False,
                },
            },
            "additionalProperties": False,
        },
        # additional commands for scriptlets
        "scripts": {
            "type": "object",
            "properties": {
                "prep": {
                    "type": "object",
                    "properties": {
                        "pre": {
                            "type": "array",
                            "items": {
                                "type": "string",
                            },
                        },
                        "post": {
                            "type": "array",
                            "items": {
                                "type": "string",
                            },
                        },
                    },
                    "additionalProperties": False,
                },
                "build": {
                    "type": "object",
                    "properties": {
                        "pre": {
                            "type": "array",
                            "items": {
                                "type": "string",
                            },
                        },
                        "post": {
                            "type": "array",
                            "items": {
                                "type": "string",
                            },
                        },
                    },
                    "additionalProperties": False,
                },
                "install": {
                    "type": "object",
                    "properties": {
                        "pre": {
                            "type": "array",
                            "items": {
                                "type": "string",
                            },
                        },
                        "post": {
                            "type": "array",
                            "items": {
                                "type": "string",
                            },
                        },
                    },
                    "additionalProperties": False,
                },
                "check": {
                    "type": "object",
                    "properties": {
                        "pre": {
                            "type": "array",
                            "items": {
                                "type": "string",
                            },
                        },
                        "post": {
                            "type": "array",
                            "items": {
                                "type": "string",
                            },
                        },
                    },
                    "additionalProperties": False,
                },
            },
            "additionalProperties": False,
        },
        # configure which tests are run
        "tests": {
            "type": "object",
            "properties": {
                "run": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "enum": ["none", "all", "lib", "bin", "doc", "bins", "tests"],
                        "uniqueItems": True,
                    },
                },
                "skip": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                "skip-exact": {
                    "type": "boolean",
                },
                "comments": {
                    "type": "array",
                    "items": {
                        "type": "string",
                    },
                },
            },
            "additionalProperties": False,
            "required": ["comments"],
        },
        # feature flags passed to cargo
        "features": {
            "type": "object",
            "properties": {
                # pass "--all-features"
                "enable-all": {
                    "type": "boolean",
                },
                # pass "--features foo,bar"
                "enable": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                # pass "--no-default-features"
                "disable-default": {
                    "type": "boolean",
                },
                # hide feature subpackages
                "hide": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
            },
            "additionalProperties": False,
        },
        # additional RPM Build/Requires
        "requires": {
            "type": "object",
            "properties": {
                # additional BuildRequires
                "build": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                # additional BuildRequires gated by the "check" bconf
                "test": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                # additional Requires for the -devel subpackage
                "lib": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                # additional Requires for the binary subpackage
                "bin": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                # additional Requires for feature subpackages
                "features": {
                    "type": "object",
                    "patternProperties": {
                        "^.*$": {
                            "type": "array",
                            "items": {
                                "type": "string",
                                "uniqueItems": True,
                            },
                        },
                    },
                    "additionalProperties": False,
                },
            },
            "additionalProperties": False,
        },
    },
    "additionalProperties": False,
}


def to_list(s, sort=True):
    if not s:
        return []
    if sort:
        return list(sorted(filter(None, (elem.strip() for elem in s.splitlines()))))
    else:
        return list(filter(None, (elem.strip() for elem in s.splitlines())))


class ConfError(ValueError):
    pass


class IniConf:
    def __init__(
        self,
        *,
        summary: Optional[str] = None,
        supported_arches: Optional[list[str]] = None,
        all_features: Optional[bool] = None,
        unwanted_features: Optional[list[str]] = None,
        enabled_features: Optional[list[str]] = None,
        buildrequires: Optional[list[str]] = None,
        testrequires: Optional[list[str]] = None,
        bin_requires: Optional[list[str]] = None,
        lib_requires: Optional[dict[Optional[str], list[str]]] = None,
        conf_file: bool = False,
    ):
        self.summary: Optional[str] = summary
        self.supported_arches: list[str] = supported_arches or list()
        self.all_features: Optional[bool] = all_features
        self.unwanted_features: list[str] = unwanted_features or list()
        self.enabled_features: list[str] = enabled_features or list()
        self.buildrequires: list[str] = buildrequires or list()
        self.testrequires: list[str] = testrequires or list()
        self.bin_requires: list[str] = bin_requires or list()
        self.lib_requires: dict[Optional[str], list[str]] = lib_requires or dict()
        self.conf_file: bool = conf_file

    def __eq__(self, other):
        if not isinstance(other, IniConf):
            return False  # pragma nocover

        # order of items in the list is consistent:
        # lists are sorted when parsing rust2rpm.conf files
        return (
            self.summary == other.summary
            and self.supported_arches == other.supported_arches
            and self.all_features == other.all_features
            and self.unwanted_features == other.unwanted_features
            and self.enabled_features == other.enabled_features
            and self.buildrequires == other.buildrequires
            and self.testrequires == other.testrequires
            and self.bin_requires == other.bin_requires
            and self.lib_requires == other.lib_requires
        )

    @staticmethod
    def load(filenames: list[str], target: str, features: set[str]) -> "IniConf":
        conf = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation())
        confs = conf.read(filenames)

        if len(confs) > 1:  # pragma nocover
            raise FileExistsError

        # clean up configuration files with deprecated names
        if ".rust2rpm.conf" in confs:  # pragma nocover
            os.rename(".rust2rpm.conf", "rust2rpm.conf")
            log.info("Renamed deprecated, hidden .rust2rpm.conf file to rust2rpm.conf.")

        if "_rust2rpm.conf" in confs:  # pragma nocover
            os.rename("_rust2rpm.conf", "rust2rpm.conf")
            log.info("Renamed deprecated _rust2rpm.conf file to rust2rpm.conf.")

        # merge target-specific configuration with default configuration
        if target not in conf:
            conf.add_section(target)
        merged = conf[target]

        # validate configuration file
        valid_targets = ["fedora", "mageia", "opensuse", "plain"]
        valid_keys = [
            "summary",
            "supported-arches",
            "all-features",
            "unwanted-features",
            "enabled-features",
            "buildrequires",
            "testrequires",
            "lib.requires",
            "bin.requires",
        ]

        # the "default" feature is always implicitly defined
        if "default" not in features:
            features.add("default")

        for feature in features:
            valid_keys.append(f"lib+{feature}.requires")

        # check section names
        for section in conf.keys():
            if section != "DEFAULT" and section not in valid_targets:
                raise ConfError(f"Invalid section: {section!r}")

        # check setting keys
        for key in merged.keys():
            if key not in valid_keys:
                raise ConfError(f"Invalid key: {key!r}")

        # parse configuration and validate settings
        settings = dict()

        if summary := merged.get("summary"):
            settings["summary"] = summary

        if supported_arches := merged.get("supported-arches"):
            settings["supported_arches"] = to_list(supported_arches, False)

        if merged.get("all-features"):
            all_features = merged.getboolean("all-features")
            settings["all_features"] = all_features

        if unwanted_features := merged.get("unwanted-features"):
            settings["unwanted_features"] = to_list(unwanted_features)

            for unwanted_feature in settings["unwanted_features"]:
                if unwanted_feature not in features:
                    raise ConfError(f'Unrecognized "unwanted" feature: {unwanted_feature!r}')

        if enabled_features := merged.get("enabled-features"):
            settings["enabled_features"] = to_list(enabled_features)

            for enabled_feature in settings["enabled_features"]:
                if enabled_feature not in features:
                    raise ConfError(f'Unrecognized "enabled" feature: {enabled_feature!r}')

        if buildrequires := merged.get("buildrequires"):
            settings["buildrequires"] = to_list(buildrequires)
        if testrequires := merged.get("testrequires"):
            settings["testrequires"] = to_list(testrequires)
        if bin_requires := merged.get("bin.requires"):
            settings["bin_requires"] = to_list(bin_requires)

        if lib_requires := merged.get("lib.requires"):
            if "lib_requires" not in settings.keys():
                settings["lib_requires"] = dict()
            settings["lib_requires"][None] = to_list(lib_requires)  # type: ignore

        for feature in features:
            if lib_feature_requires := merged.get(f"lib+{feature}.requires"):
                if "lib_requires" not in settings.keys():
                    settings["lib_requires"] = dict()
                settings["lib_requires"][feature] = to_list(lib_feature_requires)  # type: ignore

        settings["conf_file"] = len(confs) == 1

        return IniConf(**settings)  # type: ignore


def conf_comments_to_spec_comments(comments: Optional[list[str]]) -> list[str]:
    if not comments:
        return []

    lines = []
    for comment in comments:
        lines.extend(
            textwrap.wrap(
                comment,
                width=80,
                initial_indent="# * ",
                subsequent_indent="#   ",
                break_long_words=False,
                break_on_hyphens=False,
            )
        )

    return lines


class SourceOrPatch:
    def __init__(self, data: dict):
        self._data = data

    @property
    def file(self) -> str:
        return self._data["file"]

    @property
    def number(self) -> int:
        return self._data["number"]

    @property
    def comments(self) -> list[str]:
        return self._data.get("comments") or list()

    @property
    def comment_lines(self) -> list[str]:
        return conf_comments_to_spec_comments(self.comments)

    @property
    def whitespace(self) -> str:
        return " " * (16 - (len("Source") + len(str(self.number)) + 1))


class TomlConf:
    def __init__(self, data: Optional[dict] = None):
        self._data = data or dict()

    def __repr__(self) -> str:
        return repr(self._data)

    def __eq__(self, other) -> bool:
        if not isinstance(other, TomlConf):
            return False  # pragma nocover

        return self._data == other._data

    @property
    def _package(self) -> Optional[dict]:
        return self._data.get("package")

    @property
    def package_summary(self) -> Optional[str]:
        if package := self._package:
            return package.get("summary")
        else:
            return None

    @property
    def package_description(self) -> Optional[str]:
        if package := self._package:
            return package.get("description")
        else:
            return None

    @property
    def package_url(self) -> Optional[str]:
        if package := self._package:
            return package.get("url")
        else:
            return None

    @property
    def package_source_url(self) -> Optional[str]:
        if package := self._package:
            return package.get("source-url")
        else:
            return None

    @property
    def package_supported_arches(self) -> Optional[list[str]]:
        if package := self._package:
            return package.get("supported-arches")
        else:
            return None

    @property
    def package_bin_package_name(self) -> Optional[str]:
        if package := self._package:
            return package.get("bin-package-name")
        else:
            return None

    @property
    def package_cargo_install_bin(self) -> Optional[bool]:
        if package := self._package:
            return package.get("cargo-install-bin")
        else:
            return None

    @property
    def package_cargo_install_lib(self) -> Optional[bool]:
        if package := self._package:
            return package.get("cargo-install-lib")
        else:
            return None

    @property
    def package_debuginfo_level(self) -> Optional[int]:
        if package := self._package:
            return package.get("debuginfo-level")
        else:
            return None

    @property
    def package_cargo_toml_patch_comments(self) -> Optional[list[str]]:
        if package := self._package:
            return package.get("cargo-toml-patch-comments")
        else:
            return None

    @property
    def package_cargo_toml_patch_comment_lines(self) -> Optional[list[str]]:
        return conf_comments_to_spec_comments(self.package_cargo_toml_patch_comments)

    @property
    def package_extra_sources(self) -> Optional[list[SourceOrPatch]]:
        if package := self._package:
            if sources := package.get("extra-sources"):
                return [SourceOrPatch(source) for source in sources]
            else:
                return None
        else:
            return None

    @property
    def package_extra_patches(self) -> Optional[list[SourceOrPatch]]:
        if package := self._package:
            if patches := package.get("extra-patches"):
                return [SourceOrPatch(patch) for patch in patches]
            else:
                return None
        else:
            return None

    @property
    def package_extra_files(self) -> Optional[list[str]]:
        if package := self._package:
            return package.get("extra-files")
        else:
            return None

    @property
    def package_bin_renames(self) -> Optional[dict[str, str]]:
        if package := self._package:
            return package.get("bin-renames")
        else:
            return None

    @property
    def _scripts(self) -> Optional[dict[str, dict[str, list[str]]]]:
        return self._data.get("scripts")

    @property
    def _scripts_prep(self) -> Optional[dict[str, list[str]]]:
        if scripts := self._scripts:
            return scripts.get("prep")
        else:
            return None

    @property
    def scripts_prep_pre(self) -> Optional[list[str]]:
        if scripts_prep := self._scripts_prep:
            return scripts_prep.get("pre")
        else:
            return None

    @property
    def scripts_prep_post(self) -> Optional[list[str]]:
        if scripts_prep := self._scripts_prep:
            return scripts_prep.get("post")
        else:
            return None

    @property
    def _scripts_build(self) -> Optional[dict[str, list[str]]]:
        if scripts := self._scripts:
            return scripts.get("build")
        else:
            return None

    @property
    def scripts_build_pre(self) -> Optional[list[str]]:
        if scripts_build := self._scripts_build:
            return scripts_build.get("pre")
        else:
            return None

    @property
    def scripts_build_post(self) -> Optional[list[str]]:
        if scripts_build := self._scripts_build:
            return scripts_build.get("post")
        else:
            return None

    @property
    def _scripts_install(self) -> Optional[dict[str, list[str]]]:
        if scripts := self._scripts:
            return scripts.get("install")
        else:
            return None

    @property
    def scripts_install_pre(self) -> Optional[list[str]]:
        if scripts_install := self._scripts_install:
            return scripts_install.get("pre")
        else:
            return None

    @property
    def scripts_install_post(self) -> Optional[list[str]]:
        if scripts_install := self._scripts_install:
            return scripts_install.get("post")
        else:
            return None

    @property
    def _scripts_check(self) -> Optional[dict[str, list[str]]]:
        if scripts := self._scripts:
            return scripts.get("check")
        else:
            return None

    @property
    def scripts_check_pre(self) -> Optional[list[str]]:
        if scripts_check := self._scripts_check:
            return scripts_check.get("pre")
        else:
            return None

    @property
    def scripts_check_post(self) -> Optional[list[str]]:
        if scripts_check := self._scripts_check:
            return scripts_check.get("post")
        else:
            return None

    @property
    def _tests(self) -> Optional[dict[str, Any]]:
        return self._data.get("tests")

    @property
    def tests_run(self) -> Optional[list[str]]:
        if tests := self._tests:
            return tests.get("run")
        return None

    @property
    def tests_skip(self) -> Optional[list[str]]:
        if tests := self._tests:
            return tests.get("skip")
        return None

    @property
    def tests_skip_exact(self) -> Optional[bool]:
        if tests := self._tests:
            return tests.get("skip-exact")
        return None

    @property
    def tests_comments(self) -> Optional[list[str]]:
        if tests := self._tests:
            return tests.get("comments")
        return None

    @property
    def tests_comment_lines(self) -> Optional[list[str]]:
        return conf_comments_to_spec_comments(self.tests_comments)

    @property
    def _features(self) -> Optional[dict]:
        return self._data.get("features")

    @property
    def features_enable_all(self) -> Optional[bool]:
        if features := self._features:
            return features.get("enable-all")
        else:
            return None

    @property
    def features_enable(self) -> Optional[list[str]]:
        if features := self._features:
            return features.get("enable")
        else:
            return None

    @property
    def features_disable_default(self) -> Optional[bool]:
        if features := self._features:
            return features.get("disable-default")
        else:
            return None

    @property
    def features_hide(self) -> Optional[list[str]]:
        if features := self._features:
            return features.get("hide")
        else:
            return None

    @property
    def _requires(self) -> Optional[dict]:
        return self._data.get("requires")

    @property
    def requires_build(self) -> Optional[list[str]]:
        if requires := self._requires:
            return requires.get("build")
        else:
            return None

    @property
    def requires_test(self) -> Optional[list[str]]:
        if requires := self._requires:
            return requires.get("test")
        else:
            return None

    @property
    def requires_lib(self) -> Optional[list[str]]:
        if requires := self._requires:
            return requires.get("lib")
        else:
            return None

    @property
    def requires_bin(self) -> Optional[list[str]]:
        if requires := self._requires:
            return requires.get("bin")
        else:
            return None

    @property
    def requires_features(self) -> Optional[dict[str, list[str]]]:
        if requires := self._requires:
            return requires.get("features")
        else:
            return None

    @staticmethod
    def load(path: str, features: set[str]) -> "TomlConf":
        with open(path, "rb") as file:
            toml = tomllib.load(file)

        jsonschema.validate(toml, TOML_SCHEMA)

        conf = TomlConf(toml)

        # the "default" feature is always implicitly defined
        if "default" not in features:
            features.add("default")

        # validate the "tests.run" setting:
        # both "all" and "none" cannot be combimed with other values
        if tests_run := conf.tests_run:
            if "all" in tests_run:
                if len(tests_run) != 1:
                    raise ConfError(f"Invalid set of tests to run: {tests_run!r}")
            if "none" in tests_run:
                if len(tests_run) != 1:
                    raise ConfError(f"Invalid set of tests to run: {tests_run!r}")

        # validate the "features.enable-all" and "features.disable-default" settings:
        # both cannot be enabled at the same time
        if conf.features_enable_all is True and conf.features_disable_default is True:
            raise ConfError("Conflicting settings for features: 'enable-all' and 'disable-default'")

        # validate the "features.enable" setting:
        # list elements must be valid feature names
        if enable_list := conf.features_enable:
            for enabled in enable_list:
                if enabled not in features:
                    raise ConfError(f'Unrecognized "enabled" feature: {enabled!r}')

        # validate the "features.hide" setting:
        # list elements must be valid feature names
        if hide_list := conf.features_hide:
            for hidden in hide_list:
                if hidden not in features:
                    raise ConfError(f'Unrecognized "hidden" feature: {hidden!r}')

        # validate the "requires.features" setting
        # dictionary keys must be valid feature names
        if feature_requires := conf.requires_features:
            for feature in feature_requires.keys():
                if feature not in features:
                    raise ConfError(f'Unrecognized "Requires" for feature: {feature!r}')

        # warn when conflicting settings are used
        if (conf.features_enable_all is True) and (conf.features_hide is not None and len(conf.features_hide) > 0):
            log.warn(
                "Conflicting settings for features: "
                "All features are enabled for the build but some feature subpackages are hidden. "
                "This is likely an error."
            )

        return conf

    @staticmethod
    def from_conf(conf: IniConf) -> "TomlConf":
        settings: dict[str, Any] = dict()

        package: dict[str, Any] = dict()
        if conf.summary:
            package["summary"] = conf.summary
        if conf.supported_arches:
            package["supported-arches"] = conf.supported_arches
        if package:
            settings["package"] = package

        features: dict[str, Any] = dict()
        if conf.all_features is not None:
            features["enable-all"] = conf.all_features
        if conf.enabled_features:
            features["enable"] = conf.enabled_features
        if conf.unwanted_features:
            features["hide"] = conf.unwanted_features
        if features:
            settings["features"] = features

        requires: dict[str, Any] = dict()
        if conf.buildrequires:
            requires["build"] = conf.buildrequires
        if conf.testrequires:
            requires["test"] = conf.testrequires
        if conf.bin_requires:
            requires["bin"] = conf.bin_requires
        if conf.lib_requires:
            for key, value in conf.lib_requires.items():
                if key is None:
                    requires["lib"] = value
                else:
                    if "features" not in requires.keys():
                        requires["features"] = dict()
                    requires["features"][key] = value
        if requires:
            settings["requires"] = requires

        jsonschema.validate(settings, TOML_SCHEMA)

        return TomlConf(settings)

    def to_bcond_check(self) -> bool:
        if tests_run := self.tests_run:
            return "none" not in tests_run
        else:
            return True

    def to_feature_flags(self) -> FeatureFlags:
        return FeatureFlags(
            all_features=self.features_enable_all is True,
            no_default_features=self.features_disable_default is True,
            features=list(sorted(set(self.features_enable) if self.features_enable else set())),
        )

    def to_cargo_test_args(self) -> list[str]:
        if skip := self.tests_skip:
            skip_flags = []
            for name in skip:
                skip_flags.extend(["--skip", name])
        else:
            skip_flags = []

        if self.tests_skip_exact is True:
            skip_exact_flag = ["--exact"]
        else:
            skip_exact_flag = []

        if run := self.tests_run:
            if not ("all" in run or "none" in run):
                multi = []
                for kind in run:
                    if skip_flags:
                        multi.append(" " + " ".join(["--", f"--{kind}", "--"] + skip_exact_flag + skip_flags))
                    else:
                        multi.append(" " + " ".join(["--", f"--{kind}"]))
                return multi

        if skip_flags:
            return [" " + " ".join(["--", "--"] + skip_exact_flag + skip_flags)]
        else:
            return []


def load_config(features: set[str], target: str) -> TomlConf:
    # attempt to load rust2rpm.toml
    try:
        tomlconf = TomlConf.load("rust2rpm.toml", features)

    except FileNotFoundError as _exc:
        tomlconf = None

    except tomllib.TOMLDecodeError as exc:
        log.error("Cannot read rust2rpm.toml file (TOML syntax error):")
        log.error(str(exc))
        sys.exit(1)

    except jsonschema.ValidationError as exc:
        if not exc.path:
            log.error("Invalid rust2rpm.toml file (unknown setting or table):")
            log.error(exc.message)
        else:
            err_path = ""
            for elem in exc.path:
                if isinstance(elem, int):
                    err_path += f"[{elem}]"
                else:
                    err_path += f"/{elem}"
            log.error(f"Invalid rust2rpm.toml file (invalid setting at {err_path!r}):")
            log.error(exc.message)
        sys.exit(1)

    except ConfError as exc:
        log.error("Invalid rust2rpm.toml file:")
        log.error(str(exc))
        sys.exit(1)

    if tomlconf:
        return tomlconf

    # fall back to rust2rpm.conf and translate
    try:
        # known file names (only in the current working directory)
        filenames = [".rust2rpm.conf", "_rust2rpm.conf", "rust2rpm.conf"]
        distconf = IniConf.load(filenames, target, features)

    except FileExistsError:
        log.error(
            "More than one *rust2rpm.conf file is present in this directory. "
            + "Ensure that there is only one, and that it has the correct contents."
        )
        sys.exit(1)

    except ConfError as exc:
        log.error("Invalid rust2rpm configuration file:")
        log.error(str(exc))
        sys.exit(1)

    if distconf.conf_file:
        log.warn(
            "The rust2rpm.conf file format is deprecated and support for it will "
            + "be removed in a future release of rust2rpm. Please migrate settings "
            + "to the new TOML based format (rust2rpm.toml)."
        )

    tomlconf = TomlConf.from_conf(distconf)
    return tomlconf
