from typing import Optional
from urllib.parse import urljoin

import requests


DIST_GIT_API_URL = "https://src.fedoraproject.org/api/0/"


def get_package_info(package: str) -> Optional[dict]:
    """Download information about package from dist-git.

    Returns JSON with package metadata, or None if the package is
    unnkown or the spec file is not present.

    >>> get_package_info('rust-alacritty')
    {...
    'name': 'rust-alacritty',
    'namespace': 'rpms',
    ...}
    """

    # check if the dist-git repository exists

    url = urljoin(DIST_GIT_API_URL, f"rpms/{package}")
    req = requests.get(url, headers={"User-Agent": "rust2rpm"})
    data = req.json()

    if "name" not in data.keys():
        return None

    # check if there is a spec file on the rawhide branch

    full_url = data["full_url"]
    if not full_url.endswith("/"):
        # ensure the base URL has a trailing slash;
        # otherwise urljoin mangles the URL
        full_url += "/"

    spec_url = urljoin(full_url, f"blob/rawhide/f/{package}.spec")
    req = requests.head(spec_url, headers={"User-Agent": "rust2rpm"})

    if not req.ok:
        return None

    return data
