"""
Functionality for rendering the spec file template.
"""

import time
from typing import Optional

from cargo2rpm.metadata import FeatureFlags, Metadata, Package
from cargo2rpm.semver import Version
from cargo2rpm import rpm
import jinja2

from rust2rpm import __version__, log
from rust2rpm.conf import TomlConf
from rust2rpm.licensing import translate_license
from rust2rpm.metadata import package_uses_rust_1_60_feature_syntax, get_required_features_for_binaries


RUST_PACKAGING_DEPS: dict[int, str] = {
    21: "rust-packaging >= 21",
    23: "rust-packaging >= 23",
    24: "cargo-rpm-macros >= 24",
    25: "cargo-rpm-macros >= 25",
    26: "cargo-rpm-macros >= 26",
}


RUST_PACKAGING_TARGET_MIN: dict[str, int] = {
    "fedora": 24,
    "mageia": 21,
    "opensuse": 21,
    "plain": 21,
}


def license_is_composite(license: str) -> bool:
    return (
        " " in license
        and " WITH " not in license
        and "(" not in license
        and ")" not in license
        or "(" in license
        or ")" in license
    )


def min_rust_packaging_dep(
    package: Package,
    target: str,
    is_bin: bool,
    is_cdylib: bool,
    vendor_tarball: Optional[str],
    cargo_install_lib: bool,
    cargo_install_bin: bool,
) -> int:
    if (not cargo_install_lib) or (not cargo_install_bin):
        min_dep = 26
    elif vendor_tarball:
        min_dep = 25
    elif package_uses_rust_1_60_feature_syntax(package.features):
        min_dep = 24
    elif is_bin or is_cdylib:
        min_dep = 23
    else:
        min_dep = 21

    return max((min_dep, RUST_PACKAGING_TARGET_MIN[target]))


def renames_from_bin_name_map(bin_name_map: Optional[dict[str, str]]) -> list[str]:
    if bin_name_map is None:
        return []

    return [
        f"mv %{{buildroot}}/%{{_bindir}}/{old} %{{buildroot}}/%{{_bindir}}/{new}" for old, new in bin_name_map.items()
    ]


def cargo_args_from_flags(
    feature_flags: FeatureFlags, required_features: set[str], enabled_by_default: set[str]
) -> str:
    if feature_flags.all_features:
        cargo_args = " -a"

    elif required_features:
        # remove required features that are already part of the enabled feature set
        for f in enabled_by_default:
            if f in required_features:
                required_features.remove(f)

        # merge, de-duplicate, and re-sort lists of enabled features
        if features_enable := feature_flags.features:
            enabled_features = sorted(set.union(set(required_features), set(features_enable)))
        else:
            enabled_features = sorted(required_features)

        if feature_flags.no_default_features:
            cargo_args_list = ["-n"]
        else:
            cargo_args_list = []

        if len(enabled_features) > 0:
            cargo_args_list.append(f"-f {','.join(enabled_features)}")

        if cargo_args_list:
            cargo_args = " " + " ".join(cargo_args_list)
        else:
            cargo_args = ""

    elif feature_flags.features:
        # list of features is already sorted when parsing the configuration file
        if feature_flags.no_default_features:
            cargo_args_list = ["-n"]
        else:
            cargo_args_list = []

        if len(feature_flags.features) > 0:
            cargo_args_list.append(f"-f {','.join(feature_flags.features)}")

        if cargo_args_list:
            cargo_args = " " + " ".join(cargo_args_list)
        else:
            cargo_args = ""

    elif feature_flags.no_default_features:
        cargo_args = " -n"

    else:
        cargo_args = ""

    return cargo_args


def spec_file_template(filename: str):
    env = jinja2.Environment(
        loader=jinja2.ChoiceLoader([jinja2.FileSystemLoader(["/"]), jinja2.PackageLoader("rust2rpm", "templates")]),
        extensions=["jinja2.ext.do"],
        trim_blocks=True,
        lstrip_blocks=True,
    )

    return env.get_template(filename)


def template_args_fedora(date: Optional[time.struct_time], packager: Optional[str], rpmautospec: bool) -> dict:
    kwargs: dict[str, str | bool] = dict()

    kwargs["include_build_requires"] = False
    kwargs["include_provides"] = False
    kwargs["include_requires"] = False

    if rpmautospec:
        kwargs["rpm_release"] = "%autorelease"
    else:
        kwargs["rpm_release"] = "1%{?dist}"

    rpm_changelog_date_format = "%a %b %d %Y"
    if date:
        kwargs["rpm_changelog_date"] = time.strftime(rpm_changelog_date_format, date)
    else:
        kwargs["rpm_changelog_date"] = time.strftime(rpm_changelog_date_format)

    kwargs["rpm_changelog_packager"] = packager or "rust2rpm <nobody@fedoraproject.org>"

    return kwargs


def template_args_mageia(date: Optional[time.struct_time], packager: Optional[str]) -> dict:
    kwargs: dict[str, str | bool] = dict()

    kwargs["include_build_requires"] = True
    kwargs["include_provides"] = False
    kwargs["include_requires"] = False

    kwargs["rpm_release"] = "%mkrel 1"
    kwargs["rpm_group"] = "Development/Rust"

    rpm_changelog_date_format = "%a %b %d %Y"
    if date:
        kwargs["rpm_changelog_date"] = time.strftime(rpm_changelog_date_format, date)
    else:
        kwargs["rpm_changelog_date"] = time.strftime(rpm_changelog_date_format)

    kwargs["rpm_changelog_packager"] = packager or "rust2rpm <nobody@mageia.org>"

    return kwargs


def template_args_opensuse(date: Optional[time.struct_time], packager: Optional[str]) -> dict:
    kwargs: dict[str, str | bool] = dict()

    kwargs["include_build_requires"] = True
    kwargs["include_provides"] = False
    kwargs["include_requires"] = False

    kwargs["spec_copyright_year"] = time.strftime("%Y")
    kwargs["rpm_release"] = "0"
    kwargs["rpm_group"] = "Development/Libraries/Rust"

    rpm_changelog_date_format = "%a %b %d %T %Z %Y"
    if date:
        kwargs["rpm_changelog_date"] = time.strftime(rpm_changelog_date_format, date)
    else:
        kwargs["rpm_changelog_date"] = time.strftime(rpm_changelog_date_format)

    kwargs["rpm_changelog_packager"] = packager or "rust2rpm <opensuse-packaging@opensuse.org>"

    return kwargs


def template_args_plain(date: Optional[time.struct_time], packager: Optional[str]) -> dict:
    kwargs: dict[str, str | bool] = dict()

    kwargs["include_build_requires"] = True
    kwargs["include_provides"] = True
    kwargs["include_requires"] = True

    kwargs["rpm_release"] = "1%{?dist}"

    rpm_changelog_date_format = "%a %b %d %Y"
    if date:
        kwargs["rpm_changelog_date"] = time.strftime(rpm_changelog_date_format, date)
    else:
        kwargs["rpm_changelog_date"] = time.strftime(rpm_changelog_date_format)

    kwargs["rpm_changelog_packager"] = packager or "rust2rpm <packager@example.com>"

    return kwargs


def spec_render_crate(
    *,
    metadata: Metadata,
    upstream_version: str,
    target: str,
    legacy: bool,
    rpm_name: str,
    compat: bool,
    patch_file_automatic: Optional[str],
    patch_file_manual: Optional[str],
    license_files: list[str],
    doc_files: list[str],
    vendor_tarball: Optional[str],
    tomlconf: TomlConf,
    relative_license_paths: bool,
    rpmautospec: bool,
    auto_changelog_entry: bool,
    date: Optional[time.struct_time] = None,
    packager: Optional[str],
) -> str:
    template = spec_file_template("crate.spec")

    # enforce consistent ordering of binaries
    if bin_renames := tomlconf.package_bin_renames:
        binaries = []
        for old_name in metadata.get_binaries():
            if new_name := bin_renames.get(old_name):
                binaries.append(new_name)
            else:
                binaries.append(old_name)
    else:
        binaries = [*metadata.get_binaries()]
    binaries.sort()

    is_bin = metadata.is_bin()
    is_lib = metadata.is_lib()
    is_cdylib = metadata.is_cdylib()

    cargo_install_lib = (tomlconf.package_cargo_install_lib is not False) and (vendor_tarball is None)
    cargo_install_bin = (tomlconf.package_cargo_install_bin is not False) and (not compat if is_bin else True)

    if tomlconf.package_bin_package_name is None:
        bin_name = "%{crate}"
    else:
        bin_name = tomlconf.package_bin_package_name

    package = metadata.packages[0]

    if rpm_summary := tomlconf.package_summary:
        if len(rpm_summary) > 80:
            log.warn("The package Summary specified in rust2rpm.toml is too long (>80 characters).")
    elif rpm_summary := package.get_summary():
        if len(rpm_summary) > 80:
            log.warn(
                "Heuristics for generating a short Summary failed. "
                "The generated Summary string is too long (>80 characters) and needs to be shortened manually. "
                "This setting can be persisted with a 'package.summary' setting in 'rust2rpm.toml'."
            )
    else:
        rpm_summary = "# FIXME"

    rpm_description = tomlconf.package_description or package.get_description()

    # enforce consistent ordering of feature subpackages
    features: list[Optional[str]] = list(sorted(package.get_feature_names()))
    if "default" in features:
        features.remove("default")
    features.insert(0, None)
    features.insert(1, "default")

    generator_version = __version__.split(".")[0]
    feature_flags = tomlconf.to_feature_flags()

    try:
        buildrequires_with_dev = rpm.buildrequires(package, feature_flags, True)
        buildrequires_without_dev = rpm.buildrequires(package, feature_flags, False)

        buildrequires = buildrequires_without_dev
        test_requires = set.difference(buildrequires_with_dev, buildrequires_without_dev)
    except ValueError:
        # some ancient Rust crates have metadata that might be considered "invalid" by today's standards,
        # causing generation of BuildRequires to fail
        raise

    rpm_buildrequires = list(sorted(buildrequires))
    rpm_test_requires = list(sorted(test_requires))
    rpm_requires = {feature: list(sorted(rpm.requires(package, feature))) for feature in features}
    rpm_provides = {feature: rpm.provides(package, feature) for feature in features}

    conf_lib_requires = dict()
    for feature in features:
        if feature is None:
            conf_key = "lib"
            conf_lib_requires[conf_key] = tomlconf.requires_lib or list()
        else:
            conf_key = f"lib+{feature}"
            if requires_features := tomlconf.requires_features:
                conf_lib_requires[conf_key] = requires_features.get(feature) or list()
            else:
                conf_lib_requires[conf_key] = list()

    if features_hide := tomlconf.features_hide:
        for feature in features_hide:
            features.remove(feature)

    rust_packaging_dep = RUST_PACKAGING_DEPS[
        min_rust_packaging_dep(package, target, is_bin, is_cdylib, vendor_tarball, cargo_install_lib, cargo_install_bin)
    ]

    required_features = get_required_features_for_binaries(package)
    features_enabled_by_default = package.get_enabled_features_transitive(feature_flags)[0]
    cargo_args = cargo_args_from_flags(feature_flags, required_features, features_enabled_by_default)

    if license_str := metadata.packages[0].license:
        rpm_license, rpm_license_comments = translate_license(target, license_str)
    else:
        log.warn("No license information in crate metadata.")
        rpm_license = None
        rpm_license_comments = "# FIXME: No license information in crate metadata."

    if supported_arches := tomlconf.package_supported_arches:
        conf_supported_arches = " ".join(supported_arches)
    else:
        conf_supported_arches = None

    rpm_bcond_check = tomlconf.to_bcond_check()
    rpm_test_comments = tomlconf.tests_comment_lines
    cargo_test_args = tomlconf.to_cargo_test_args()

    template_args_common = {
        # Parameters specific to rust2rpm
        "rust2rpm_version": generator_version,
        "rust2rpm_target": target,
        "rust2rpm_legacy": legacy,
        "rust_packaging_dep": rust_packaging_dep,
        # Paramters that control compiler flags
        "build_rustflags_debuginfo": tomlconf.package_debuginfo_level,
        # Parameters for RPM package metadata
        "rpm_name": rpm_name,
        "rpm_version": Version.parse(package.version).to_rpm(),
        "rpm_summary": rpm_summary,
        "rpm_description": rpm_description,
        "rpm_license": rpm_license,
        "rpm_license_comments": rpm_license_comments,
        "rpm_patch_file_automatic": patch_file_automatic,
        "rpm_patch_file_manual": patch_file_manual,
        "rpm_patch_file_comments": tomlconf.package_cargo_toml_patch_comment_lines,
        "rpm_buildrequires": rpm_buildrequires,
        "rpm_test_requires": rpm_test_requires,
        "rpm_requires": rpm_requires,
        "rpm_provides": rpm_provides,
        "rpm_license_files": license_files,
        "rpm_doc_files": doc_files,
        "rpm_bcond_check": rpm_bcond_check,
        "rpm_test_comments": rpm_test_comments,
        "rpm_vendor_source": vendor_tarball,
        "rpm_extra_sources": tomlconf.package_extra_sources or list(),
        "rpm_extra_patches": tomlconf.package_extra_patches or list(),
        "rpm_extra_files": tomlconf.package_extra_files or list(),
        # Parameters that control generation of subpackages
        "rpm_binary_package": is_bin,
        "rpm_binary_package_name": bin_name,
        "rpm_cdylib_package": is_cdylib,
        "rpm_library_package": is_lib,
        "rpm_binary_names": binaries,
        "crate_features": features,
        "cargo_install_lib": cargo_install_lib,
        "cargo_install_bin": cargo_install_bin,
        # Parameters that allow injecting additional commands into scriptlets
        "rpm_prep_pre": tomlconf.scripts_prep_pre or list(),
        "rpm_prep_post": tomlconf.scripts_prep_post or list(),
        "rpm_build_pre": tomlconf.scripts_build_pre or list(),
        "rpm_build_post": tomlconf.scripts_build_post or list(),
        "rpm_install_pre": tomlconf.scripts_install_pre or list(),
        "rpm_install_post": tomlconf.scripts_install_post or list(),
        "rpm_check_pre": tomlconf.scripts_check_pre or list(),
        "rpm_check_post": tomlconf.scripts_check_post or list(),
        # Parameters for crate metadata
        "crate_name": package.name,
        "crate_version": package.version,
        "crate_license": package.license,
        "upstream_version": upstream_version,
        # Parameters for RPM macros
        "rpm_autosetup_args": " -a1" if vendor_tarball else "",
        "cargo_args": cargo_args,
        "cargo_prep_args": " -v vendor" if vendor_tarball else "",
        "cargo_test_args": cargo_test_args,
        # Parameters derived from rust2rpm.toml
        "conf_buildrequires": tomlconf.requires_build or list(),
        "conf_test_requires": tomlconf.requires_test or list(),
        "conf_bin_requires": tomlconf.requires_bin or list(),
        "conf_lib_requires": conf_lib_requires,
        "conf_supported_arches": conf_supported_arches,
        "rpm_bin_renames": renames_from_bin_name_map(tomlconf.package_bin_renames),
        # Parameters derived from command-line flags
        "use_vendor_tarball": vendor_tarball is not None,
        "use_relative_license_paths": relative_license_paths,
        "use_rpmautospec": rpmautospec,
        "make_changelog_entry": auto_changelog_entry,
    }

    match target:
        case "fedora":
            template_args_target = template_args_fedora(date, packager, rpmautospec)
        case "mageia":
            template_args_target = template_args_mageia(date, packager)
        case "opensuse":
            template_args_target = template_args_opensuse(date, packager)
        case "plain":
            template_args_target = template_args_plain(date, packager)
        case _:
            raise ValueError(f"Unknown target {target!r} (this should never happen)")

    spec_contents = template.render(
        **template_args_common,
        **template_args_target,
    )

    if not spec_contents.endswith("\n"):
        spec_contents += "\n"

    return spec_contents


def spec_render_project(
    *,
    metadata: Metadata,
    upstream_version: str,
    target: str,
    legacy: bool,
    rpm_name: str,
    patch_file_automatic: Optional[str],
    patch_file_manual: Optional[str],
    license_files: list[str],
    doc_files: list[str],
    vendor_tarball: Optional[str],
    tomlconf: TomlConf,
    rpmautospec: bool,
    auto_changelog_entry: bool,
    date: Optional[time.struct_time] = None,
    packager: Optional[str],
) -> str:
    template = spec_file_template("project.spec")

    # enforce consistent ordering of binaries
    if bin_renames := tomlconf.package_bin_renames:
        binaries = []
        for old_name in metadata.get_binaries():
            if new_name := bin_renames.get(old_name):
                binaries.append(new_name)
            else:
                binaries.append(old_name)
    else:
        binaries = [*metadata.get_binaries()]
    binaries.sort()

    is_bin = metadata.is_bin()
    is_cdylib = metadata.is_cdylib()

    package = metadata.packages[0]

    if rpm_summary := tomlconf.package_summary:
        if len(rpm_summary) > 80:
            log.warn("The package Summary specified in rust2rpm.toml is too long (>80 characters).")
    elif rpm_summary := package.get_summary():
        if len(rpm_summary) > 80:
            log.warn(
                "Heuristics for generating a short Summary failed. "
                "The generated Summary string is too long (>80 characters) and needs to be shortened manually. "
                "This setting can be persisted with a 'package.summary' setting in 'rust2rpm.toml'."
            )
    else:
        rpm_summary = "# FIXME"

    rpm_description = tomlconf.package_description or package.get_description()

    generator_version = __version__.split(".")[0]
    feature_flags = tomlconf.to_feature_flags()

    buildrequires = rpm.buildrequires(package, feature_flags, False)
    test_requires = set.difference(
        rpm.buildrequires(package, feature_flags, True), rpm.buildrequires(package, feature_flags, False)
    )

    rpm_buildrequires = list(sorted(buildrequires))
    rpm_test_requires = list(sorted(test_requires))

    rust_packaging_dep = RUST_PACKAGING_DEPS[
        min_rust_packaging_dep(package, target, is_bin, is_cdylib, vendor_tarball, False, True)
    ]

    required_features = get_required_features_for_binaries(package)
    features_enabled_by_default = package.get_enabled_features_transitive(feature_flags)[0]
    cargo_args = cargo_args_from_flags(feature_flags, required_features, features_enabled_by_default)

    if license_str := metadata.packages[0].license:
        rpm_license, rpm_license_comments = translate_license(target, license_str)
    else:
        log.warn("No license information in crate metadata.")
        rpm_license = None
        rpm_license_comments = "# FIXME: No license information in crate metadata."

    if supported_arches := tomlconf.package_supported_arches:
        conf_supported_arches = " ".join(supported_arches)
    else:
        conf_supported_arches = None

    rpm_bcond_check = tomlconf.to_bcond_check()
    rpm_test_comments = tomlconf.tests_comment_lines
    cargo_test_args = tomlconf.to_cargo_test_args()

    template_args_common = {
        # Parameters specific to rust2rpm
        "rust2rpm_version": generator_version,
        "rust2rpm_target": target,
        "rust2rpm_legacy": legacy,
        "rust_packaging_dep": rust_packaging_dep,
        # Paramters that control compiler flags
        "build_rustflags_debuginfo": tomlconf.package_debuginfo_level,
        # Parameters for RPM package metadata
        "rpm_name": rpm_name,
        "rpm_version": Version.parse(package.version).to_rpm(),
        "rpm_summary": rpm_summary,
        "rpm_description": rpm_description,
        "rpm_url": tomlconf.package_url or package.repository or package.homepage or "# FIXME",
        "rpm_source_url": tomlconf.package_source_url or "# FIXME",
        "rpm_license": rpm_license,
        "rpm_license_comments": rpm_license_comments,
        "rpm_patch_file_automatic": patch_file_automatic,
        "rpm_patch_file_manual": patch_file_manual,
        "rpm_patch_file_comments": tomlconf.package_cargo_toml_patch_comment_lines,
        "rpm_buildrequires": rpm_buildrequires,
        "rpm_test_requires": rpm_test_requires,
        "rpm_license_files": license_files,
        "rpm_doc_files": doc_files,
        "rpm_bcond_check": rpm_bcond_check,
        "rpm_test_comments": rpm_test_comments,
        "rpm_vendor_source": vendor_tarball,
        "rpm_extra_sources": tomlconf.package_extra_sources or list(),
        "rpm_extra_patches": tomlconf.package_extra_patches or list(),
        "rpm_extra_files": tomlconf.package_extra_files or list(),
        # Parameters that control generation of subpackages
        "rpm_binary_package": is_bin,
        "rpm_cdylib_package": is_cdylib,
        "rpm_binary_names": binaries,
        # Parameters that allow injecting additional commands into scriptlets
        "rpm_prep_pre": tomlconf.scripts_prep_pre or list(),
        "rpm_prep_post": tomlconf.scripts_prep_post or list(),
        "rpm_build_pre": tomlconf.scripts_build_pre or list(),
        "rpm_build_post": tomlconf.scripts_build_post or list(),
        "rpm_install_pre": tomlconf.scripts_install_pre or list(),
        "rpm_install_post": tomlconf.scripts_install_post or list(),
        "rpm_check_pre": tomlconf.scripts_check_pre or list(),
        "rpm_check_post": tomlconf.scripts_check_post or list(),
        # Parameters for crate metadata
        "crate_name": package.name,
        "crate_version": package.version,
        "crate_license": package.license,
        "upstream_version": upstream_version,
        # Parameters for RPM macros
        "rpm_autosetup_args": " -a1" if vendor_tarball else "",
        "cargo_args": cargo_args,
        "cargo_prep_args": " -v vendor" if vendor_tarball else "",
        "cargo_test_args": cargo_test_args,
        # Parameters derived from rust2rpm.toml
        "conf_buildrequires": tomlconf.requires_build or list(),
        "conf_test_requires": tomlconf.requires_test or list(),
        "conf_bin_requires": tomlconf.requires_bin or list(),
        "conf_supported_arches": conf_supported_arches,
        "rpm_bin_renames": renames_from_bin_name_map(tomlconf.package_bin_renames),
        # Parameters derived from command-line flags
        "use_vendor_tarball": vendor_tarball is not None,
        "use_rpmautospec": rpmautospec,
        "make_changelog_entry": auto_changelog_entry,
    }

    match target:
        case "fedora":
            template_args_target = template_args_fedora(date, packager, rpmautospec)
        case "mageia":
            template_args_target = template_args_mageia(date, packager)
        case "opensuse":
            template_args_target = template_args_opensuse(date, packager)
        case "plain":
            template_args_target = template_args_plain(date, packager)
        case _:
            raise ValueError(f"Unknown target {target!r} (this should never happen)")

    spec_contents = template.render(
        **template_args_common,
        **template_args_target,
    )

    if not spec_contents.endswith("\n"):
        spec_contents += "\n"

    return spec_contents


def spec_render_workspace(
    *,
    metadata: Metadata,
    main_package: Package,
    upstream_version: str,
    target: str,
    legacy: bool,
    rpm_name: str,
    license_files: list[str],
    doc_files: list[str],
    vendor_tarball: Optional[str],
    tomlconf: TomlConf,
    rpmautospec: bool,
    auto_changelog_entry: bool,
    date: Optional[time.struct_time] = None,
    packager: Optional[str],
) -> str:
    template = spec_file_template("workspace.spec")

    # enforce consistent ordering of binaries
    if bin_renames := tomlconf.package_bin_renames:
        binaries = []
        for old_name in metadata.get_binaries():
            if new_name := bin_renames.get(old_name):
                binaries.append(new_name)
            else:
                binaries.append(old_name)
    else:
        binaries = [*metadata.get_binaries()]
    binaries.sort()

    is_cdylib = metadata.is_cdylib()

    try:
        rpm_version = Version.parse(upstream_version).to_rpm()
    except ValueError:
        log.warn(f"Version {upstream_version!r} is not valid according to SemVer.")
        rpm_version = upstream_version

    if rpm_summary := tomlconf.package_summary:
        if len(rpm_summary) > 80:
            log.warn("The package Summary specified in rust2rpm.toml is too long (>80 characters).")
    elif rpm_summary := main_package.get_summary():
        if len(rpm_summary) > 80:
            log.warn(
                "Heuristics for generating a short Summary failed. "
                "The generated Summary string is too long (>80 characters) and needs to be shortened manually. "
                "This setting can be persisted with a 'package.summary' setting in rust2rpm.toml."
            )
    else:
        rpm_summary = "# FIXME"

    rpm_description = tomlconf.package_description or main_package.get_description()

    generator_version = __version__.split(".")[0]
    feature_flags = tomlconf.to_feature_flags()

    buildrequires = rpm.workspace_buildrequires(metadata, feature_flags, False)
    test_requires = set.difference(
        rpm.workspace_buildrequires(metadata, feature_flags, True),
        rpm.workspace_buildrequires(metadata, feature_flags, False),
    )

    rpm_buildrequires = list(sorted(buildrequires))
    rpm_test_requires = list(sorted(test_requires))

    rust_packaging_dep = RUST_PACKAGING_DEPS[
        max(
            (
                24,
                max(
                    min_rust_packaging_dep(
                        package, target, package.is_bin(), package.is_cdylib(), vendor_tarball, True, True
                    )
                    for package in metadata.packages
                ),
            )
        )
    ]

    cargo_args = cargo_args_from_flags(feature_flags, set(), set())

    license_strs = list({package.license for package in metadata.packages if package.license})
    if len(license_strs) == 1:
        rpm_license_tag, rpm_license_comments = translate_license(target, license_strs[0])
    else:
        license_strs = [
            license_str if not license_is_composite(license_str) else f"({license_str})"
            for license_str in license_strs
            if license_str is not None
        ]
        license_strs.sort()
        license_str = " AND ".join(license_strs)
        rpm_license_tag, rpm_license_comments = translate_license(target, license_str)

    rpm_bcond_check = tomlconf.to_bcond_check()
    rpm_test_comments = tomlconf.tests_comment_lines
    cargo_test_args = tomlconf.to_cargo_test_args()

    template_args_common = {
        # Parameters specific to rust2rpm
        "rust2rpm_version": generator_version,
        "rust2rpm_target": target,
        "rust2rpm_legacy": legacy,
        "rust_packaging_dep": rust_packaging_dep,
        # Paramters that control compiler flags
        "build_rustflags_debuginfo": tomlconf.package_debuginfo_level,
        # Parameters for RPM package metadata
        "rpm_name": rpm_name,
        "rpm_version": rpm_version,
        "rpm_summary": rpm_summary,
        "rpm_description": rpm_description,
        "rpm_url": tomlconf.package_url or "# FIXME",
        "rpm_source_url": tomlconf.package_source_url or "# FIXME",
        "rpm_license": rpm_license_tag,
        "rpm_license_comments": rpm_license_comments,
        "rpm_buildrequires": rpm_buildrequires,
        "rpm_test_requires": rpm_test_requires,
        "rpm_license_files": license_files,
        "rpm_doc_files": doc_files,
        "rpm_binary_names": binaries,
        "rpm_cdylib_package": is_cdylib,
        "upstream_version": upstream_version,
        "rpm_bcond_check": rpm_bcond_check,
        "rpm_test_comments": rpm_test_comments,
        "rpm_vendor_source": vendor_tarball,
        "rpm_extra_sources": tomlconf.package_extra_sources or list(),
        "rpm_extra_patches": tomlconf.package_extra_patches or list(),
        "rpm_extra_files": tomlconf.package_extra_files or list(),
        # Parameters that allow injecting additional commands into scriptlets
        "rpm_prep_pre": tomlconf.scripts_prep_pre or list(),
        "rpm_prep_post": tomlconf.scripts_prep_post or list(),
        "rpm_build_pre": tomlconf.scripts_build_pre or list(),
        "rpm_build_post": tomlconf.scripts_build_post or list(),
        "rpm_install_pre": tomlconf.scripts_install_pre or list(),
        "rpm_install_post": tomlconf.scripts_install_post or list(),
        "rpm_check_pre": tomlconf.scripts_check_pre or list(),
        "rpm_check_post": tomlconf.scripts_check_post or list(),
        # Parameters for RPM macros
        "rpm_autosetup_args": " -a1" if vendor_tarball else "",
        "cargo_args": cargo_args,
        "cargo_prep_args": " -v vendor" if vendor_tarball else "",
        "cargo_test_args": cargo_test_args,
        # Parameters derived from rust2rpm.toml
        "conf_buildrequires": tomlconf.requires_build or list(),
        "conf_test_requires": tomlconf.requires_test or list(),
        "conf_bin_requires": tomlconf.requires_bin or list(),
        "rpm_bin_renames": renames_from_bin_name_map(tomlconf.package_bin_renames),
        # Parameters derived from command-line flags
        "use_vendor_tarball": vendor_tarball is not None,
        "use_rpmautospec": rpmautospec,
        "make_changelog_entry": auto_changelog_entry,
    }

    match target:
        case "fedora":
            template_args_target = template_args_fedora(date, packager, rpmautospec)
        case "mageia":
            template_args_target = template_args_mageia(date, packager)
        case "opensuse":
            template_args_target = template_args_opensuse(date, packager)
        case "plain":
            template_args_target = template_args_plain(date, packager)
        case _:
            raise ValueError(f"Unknown target {target!r} (this should never happen)")

    spec_contents = template.render(
        **template_args_common,
        **template_args_target,
    )

    if not spec_contents.endswith("\n"):
        spec_contents += "\n"

    return spec_contents
