import contextlib
import glob
import os
from pathlib import Path
import re
import sys
import shlex
import shutil
import subprocess
import requests
from typing import Optional

from cargo2rpm.semver import Version

from rust2rpm import log


DEFAULT_EDITOR = "nano"


@contextlib.contextmanager
def remove_on_error(path):
    try:
        yield
    except:  # this is supposed to include ^C
        os.unlink(path)
        raise


@contextlib.contextmanager
def exit_on_common_errors():
    """Suppress tracebacks on common "expected" exceptions"""
    try:
        yield
    except requests.exceptions.HTTPError as e:
        sys.exit(f"Failed to download metadata: {e}")
    except subprocess.CalledProcessError as e:
        cmd = shlex.join(e.cmd)
        sys.exit(f"Subcommand failed with code {e.returncode}: {cmd}")
    except FileNotFoundError as e:
        sys.exit(str(e))


def detect_editor() -> str:
    terminal = os.getenv("TERM")
    terminal_is_dumb = not terminal or terminal == "dumb"
    editor = None
    if not terminal_is_dumb:
        editor = os.getenv("VISUAL")
    if not editor:
        editor = os.getenv("EDITOR")
    if not editor:
        if terminal_is_dumb:
            raise Exception("Terminal is dumb, but $EDITOR unset")
        editor = DEFAULT_EDITOR
    return editor


def detect_packager() -> Optional[str]:
    # If we're forcing the fallback...
    if os.getenv("RUST2RPM_NO_DETECT_PACKAGER"):
        return None

    # If we're supplying packager identity through an environment variable...
    if packager := os.getenv("RUST2RPM_PACKAGER"):
        return packager

    # If we're detecting packager identity through rpmdev-packager...
    if rpmdev_packager := shutil.which("rpmdev-packager"):
        return subprocess.check_output(rpmdev_packager, text=True).strip()

    # If we're detecting packager identity through git configuration...
    if git := shutil.which("git"):
        name = subprocess.check_output([git, "config", "user.name"], text=True).strip()
        email = subprocess.check_output([git, "config", "user.email"], text=True).strip()
        return f"{name} <{email}>"

    return None


def guess_crate_name() -> Optional[str]:
    """Guess crate name from directory name and/or spec file name

    If a spec file is present, we use the %crate variable. This is the best
    option. But if we're in a new directory, and don't have a spec file yet,
    let's assume that this is a dist-git directory and extract the crate name.
    For compat packages, the directory name will contain a version suffix,
    but a compat package would almost always be created from an existing
    dist-git directory, hence there'd be a spec file, so we can ignore this.
    """
    specs = glob.glob("*.spec")

    if len(specs) > 1:
        log.error(
            "Found multiple spec files in the current working directory; "
            + "unable to determine crate name automatically."
        )
        return None

    if len(specs) == 1:
        crate = None
        spec = specs[0]

        for line in open(spec):
            if m := re.match(r"^%(?:global|define)\s+crate\s+(\S+)\s+", line):
                if crate:
                    log.error(
                        f"Found multiple definitions of the '%crate' macro in {spec!r}; "
                        + "unable to determine crate name automatically."
                    )
                    return None
                crate = m.group(1)
                if "%" in crate:
                    log.error("The value of the %crate macro appears to contain other macros and cannot be parsed.")
                    return None
        if crate:
            log.success(f"Found valid spec file {spec!r} for the {crate!r} crate.")
        else:
            log.error(f"Invalid spec file {spec!r}; unable to determine crate name automatically.")
        return crate

    dirname = os.path.basename(os.getcwd())
    if m := re.match("^rust-([a-z+0-9_-]+)$", dirname):
        crate = m.group(1)
        log.info(f"Continuing with crate name {crate!r} based on the current working directory.")
        return crate

    return None


def package_name_suffixed(name: str, suffix: Optional[str]):
    joiner = "_" if suffix and name[-1].isdigit() else ""
    return "rust-" + name + joiner + (suffix or "")


def package_name_compat(name: str, version: str):
    sv = Version.parse(version)
    if sv.major > 0:
        suffix = str(sv.major)
    elif sv.minor > 0:
        suffix = f"0.{str(sv.minor)}"
    else:
        suffix = f"0.0.{str(sv.patch)}"
    return package_name_suffixed(name, suffix)


def detect_rpmautospec(default_target: str, spec_file: Path):
    """Guess whether %autorelease+%autochangelog should be used

    Returns False if we're not on Fedora or if the spec file exists and
    wasn't using rpmautospec already.
    """

    # We default to on only for selected distros for now…
    if default_target not in {"fedora"}:
        return False

    try:
        text = spec_file.read_text()
    except FileNotFoundError:
        # A new file, let's try the new thing.
        return True

    # We are redoing an existing spec file. Figure out if it had
    # %autochangelog enabled before.
    autochangelog_re = r"^\s*%(?:autochangelog|\{\??autochangelog\})\b"
    return any(re.match(autochangelog_re, line) for line in text.splitlines())
